#include <JTC\jtc.h>
#include <iostream.h>

/*
  Una semplice classe matrice. Ogni elemento  accessibile in lettura e scrittura come M(i,j)
*/
class Matrice
{
	double* A;
	int M;
public:
	Matrice() : A(0), M(0) {}
	Matrice(int M)
	{ 
        this->M = M;
		A = new double[M*M];
	}
	double& operator()(int i, int j)
	{
		return A[i + j*M];
	}
	operator=(Matrice& Ma)
	{
		if (M == Ma.M)
			for(int i = 0; i < M*M; i++)
				A[i] = Ma.A[i];
	}
	void set(int j)
	{
			for(int i = 0; i < M*M; i++)
				A[i] = j;
	
	}
	~Matrice()
	{
		delete A;
	}
    friend ostream& operator<<(ostream& os, Matrice& M);

};
/*
   Operatore per la stampa di matrici in output
*/
ostream& operator<<(ostream& os, Matrice& M)
{
	for(int i=0; i < M.M; i++) {
		for(int j=0; j < M.M; j++)
			os << M(i,j) << "\t";
	    os << "\n";
	}
	return os;
}


class SfrancicaMatrice : public JTCMonitor
{
	friend class Sfrancicatore; // L'amicizia, prima di tutto

	Matrice& A; // Primo operando
	Matrice& B; // Secondo operando
	Matrice& C; // Terzo operando
	int N;      // Numero di thread
	int M;      // Lato delle matrici
	int nextAssignR;  // Contiene l'indice di riga della prossima cella da assegnare
	int nextAssignC;  // Contiene l'indice di colonna della prossima cella da assegnare
	int assignedCells; // Conteggia il numero di celle gi assegnate a un thread
	int threadsFinished; // Conteggia il numero di thread che hanno terminato
    
	double prodottoScalare(int r, int c); // Calcola il prodotto scalare tra una data riga di A e una colonna di B
	bool isComplete(); // vero se tutti i thread hanno finito (threadsFinished == N)
	void waitForCompletion(); // invocata per attendere che tutti i thread terminino
	bool assignNextRC(int&r, int& c); // invocata per ottenere l'assegnazione di una nuova cella di C da calcolare. False se non ci
    // ci sono pi celle da calcolare.
	void setDone(); // Invocata da ogni thread al termine definitivo del proprio lavoro.
	
public:
    SfrancicaMatrice();
	SfrancicaMatrice(Matrice& A, Matrice& B, Matrice& C, int N, int M);
	SfrancicaMatrice(SfrancicaMatrice& M);
	void calcolaProdotto(); // calcolo prodotto matriciale in parallelo
};

typedef Sfrancicatore* PSfrancicatore;

class Sfrancicatore : public JTCThread
{
    SfrancicaMatrice* S;
public:
	Sfrancicatore(SfrancicaMatrice* S) : S(S)
	{}
	virtual void run()
	{
		int r,c;
		while(S->assignNextRC(r,c)) 
			S->C(r,c) = S->prodottoScalare(r,c);
		S->setDone();
	}
};

double SfrancicaMatrice::prodottoScalare(int r, int c)
{
      double totale = 0;
	  for (int i = 0; i < M; i++)
		totale += A(r,i)*B(i,c);
      return totale;
}

bool SfrancicaMatrice::isComplete()
	{
		return threadsFinished == N;
	}

    void SfrancicaMatrice::waitForCompletion()
	{
		JTCSynchronized sync(*this);
		// non  necessario un while: ci si addormenta UNA volta sola, e si viene risvegliati UNA volta sola.
		if(!isComplete()) wait();
	}

	bool SfrancicaMatrice::assignNextRC(int&r, int& c)
	{
		JTCSynchronized sync(*this);
		r = nextAssignR;
		c = nextAssignC;
		nextAssignC = (nextAssignC + 1) % M;
		if (nextAssignC == 0) 
			nextAssignR = (nextAssignR + 1) % M;
	    return ++assignedCells <= M*M;
	}
	void SfrancicaMatrice::setDone()
	{
		JTCSynchronized sync(*this);
        // l'ultimo thread che termina, esegue notify(). Il metodo chiamante (calcolaProdotto), viene cos sbloccato.
		if (++threadsFinished == N) 
			notify();
	}

	SfrancicaMatrice::SfrancicaMatrice(Matrice& A, Matrice& B, Matrice& C, int N, int M) : A(A), B(B), C(C), N(N), M(M)
	{}

	SfrancicaMatrice::SfrancicaMatrice(SfrancicaMatrice& Ma)
		: A(Ma.A),
		  B(Ma.B),
		  C(Ma.C),
		  N(Ma.N),
		  M(Ma.M),
		  nextAssignR(Ma.nextAssignC),
		  nextAssignC(Ma.nextAssignR),
		  assignedCells(Ma.assignedCells),
		  threadsFinished(Ma.threadsFinished)
	{}

    void SfrancicaMatrice::calcolaProdotto()
	{
		PSfrancicatore* sfr;
		sfr = new PSfrancicatore[N];
		// Inizializza variabili di servizio
		// Notare che non  necessario sincronizzare questa parte di codice: in questo momento c' un solo thread attivo.
		assignedCells = threadsFinished = nextAssignR = nextAssignC = 0;
		for(int i = 0; i < N; i++) 
		{
			sfr[i] = new Sfrancicatore(this);
			sfr[i]->start();
		}
		// attende diligentemente che tutti i thread si fermino.
		waitForCompletion();
	}

#define SIZE 5
#define THREADS 2

// Main di prova.

void main() { 
	JTCInitialize init;
    Matrice A(SIZE),B(SIZE),C(SIZE);
	A.set(5); B.set(3); C.set(0);
    SfrancicaMatrice S(A,B,C,THREADS,SIZE);
    S.calcolaProdotto();    
	cout << A << B << C;
 }
